/*
 * Sonar, entreprise quality control tool.
 * Copyright (C) 2007-2008 Hortis-GRC SA
 * mailto:be_agile HAT hortis DOT ch
 *
 * Sonar is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * Sonar is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with Sonar; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02
 */
package org.sonar.maven;

import ch.hortis.sonar.model.*;
import ch.hortis.sonar.service.MeasureKey;
import org.sonar.commons.MetricsRepository;
import org.sonar.commons.database.DatabaseManager;

import javax.persistence.NoResultException;
import javax.persistence.Query;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;

public class MeasuresRecorderImpl implements MeasuresRecorder {
  private MavenProject root;
  private DatabaseManager databaseManager;
  private Date snapshotDate;
  private Map<MavenProject, Snapshot> snapshotByEntity;
  private Snapshot rootSnapshot;
  private MetricsRepository metricsRepository;
  private org.apache.maven.project.MavenProject mavenProject;

  public MeasuresRecorderImpl(DatabaseManager databaseManager, MetricsRepository metricsRepository, MavenProject root, Date snapshotDate, org.apache.maven.project.MavenProject mavenProject) {
    this.root = root;
    this.databaseManager = databaseManager;
    this.snapshotDate = snapshotDate;
    this.snapshotByEntity = new HashMap<MavenProject, Snapshot>();
    this.metricsRepository = metricsRepository;
    this.mavenProject = mavenProject;
  }


  public void start() {
    MavenProject savedRoot = getMavenProject(root.getMavenGroupId(), root.getMavenArtifactId(), root.getMavenBranch());
    if (savedRoot == null) {
      root = (MavenProject) databaseManager.save(root);

    } else {
      root = savedRoot;
      root.getProjectLinks().size();
    }

    rootSnapshot = new Snapshot();
    rootSnapshot.setMavenProject(root);
    rootSnapshot.setVersion(mavenProject.getVersion());
    rootSnapshot.setCreatedAt(snapshotDate);
    databaseManager.save(rootSnapshot);
  }

  public void stop() {

  }

  protected void finalize() throws Throwable {
    super.finalize();
    stop();
  }

  public MavenProject getRoot() {
    return root;
  }

  public Snapshot getRootSnapshot() {
    return rootSnapshot;
  }

  public void persistRoot() {
    if (root != null) {
      databaseManager.merge(root);
    }
  }

  public ProjectMeasure createProjectMeasure(MeasureKey key, Double value) {
    key.setMetric(metricsRepository.getMetric(key.getMetric()));
    return (ProjectMeasure) databaseManager.save(new ProjectMeasure(rootSnapshot, key, value));
  }

  public ProjectMeasure createPackageMeasure(String packageName, MeasureKey key, Double value) {
    MavenProject packageEntity = JavaEntities.newPackage(root.getKey(), packageName, root.getId());
    Snapshot packageSnapshot = getOrCreateSnapshot(packageEntity, rootSnapshot);

    key.setMetric(metricsRepository.getMetric(key.getMetric()));
    return (ProjectMeasure) databaseManager.save(new ProjectMeasure(packageSnapshot, key, value));
  }

  public ProjectMeasure createClassMeasure(String packageName, String className, boolean unitTest, MeasureKey key, Double value) {
    Snapshot classSnapshot = getOrCreateClassSnapshot(packageName, className, unitTest);

    key.setMetric(metricsRepository.getMetric(key.getMetric()));
    return (ProjectMeasure) databaseManager.save(new ProjectMeasure(classSnapshot, key, value));
  }

  public RuleFailure createClassRuleFailure(Rule rule, String packageName, String className, boolean unitTest, String message, RuleFailureLevel level, RuleFailureParam... params) {
    Snapshot classSnapshot = getOrCreateClassSnapshot(packageName, className, unitTest);

    RuleFailure failure = new RuleFailure(rule, level);
    failure.setMessage(message);
    failure.setSnapshot(classSnapshot);
    failure = (RuleFailure) databaseManager.save(failure);

    for (RuleFailureParam param : params) {
      if (param != null) {
        param.setRuleFailure(failure);
        param.setSnapshot(classSnapshot);
        failure.getParameters().add(param);
        databaseManager.save(param);
      }
    }
    return failure;
  }

  public SnapshotSource createClassSource(String packageName, String className, String data, boolean unitTest) {
    Snapshot classSnapshot = getOrCreateClassSnapshot(packageName, className, unitTest);
    SnapshotSource source = new SnapshotSource(classSnapshot, data);
    return (SnapshotSource) databaseManager.save(source);
  }

  private Snapshot getOrCreateClassSnapshot(String packageName, String className, boolean unitTest) {
    MavenProject packageEntity = JavaEntities.newPackage(root.getKey(), packageName, root.getId());
    MavenProject classEntity = JavaEntities.newClass(root.getKey(), packageName, className, unitTest, root.getId());

    Snapshot packageSnapshot = getOrCreateSnapshot(packageEntity, rootSnapshot);
    Snapshot classSnapshot = getOrCreateSnapshot(classEntity, packageSnapshot);
    return classSnapshot;
  }

  private MavenProject getMavenProject(String groupId, String artifactId, String branch) {
    try {
      Query query = databaseManager.getEntityManager().createQuery("SELECT p FROM MavenProject p WHERE p.scope=:scope AND p.key=:key AND p.enabled=true");
      query.setParameter("scope", MavenProject.SCOPE_PROJECT);
      query.setParameter("key", MavenProject.toMavenKey(groupId, artifactId, branch));
      return (MavenProject) query.getSingleResult();

    } catch (NoResultException e) {
      return null;
    }
  }

  private MavenProject getOrCreateEntity(MavenProject entity) {
    if (entity.equals(root)) {
      return root;
    }
    try {
      String hql = "SELECT p FROM MavenProject p WHERE p.scope=:scope AND p.key=:key AND p.rootId=:rootId AND p.enabled=true";
      if (entity.getQualifier() == null) {
        hql += " AND p.qualifier IS NULL";
      } else {
        hql += " AND p.qualifier=:qualifier";
      }
      Query query = databaseManager.getEntityManager().createQuery(hql);
      query.setParameter("scope", entity.getScope());
      query.setParameter("key", entity.getKey());
      query.setParameter("rootId", root.getId());
      if (entity.getQualifier() != null) {
        query.setParameter("qualifier", entity.getQualifier());
      }
      return (MavenProject) query.getSingleResult();

    } catch (NoResultException e) {
      entity.setRootId(root.getId());
      return (MavenProject) databaseManager.save(entity);
    }
  }

  private Snapshot getOrCreateSnapshot(MavenProject entity, Snapshot parent) {
    Snapshot snapshot = snapshotByEntity.get(entity);
    if (snapshot == null) {
      snapshot = new Snapshot();
      snapshot.setMavenProject(databaseManager.reattach(MavenProject.class, getOrCreateEntity(entity).getId()));

      snapshot.setCreatedAt(snapshotDate);
      if (parent != null) {
        parent = databaseManager.reattach(Snapshot.class, parent.getId());
        snapshot.setParent(parent);

        if (parent.getRoot() == null) {
          snapshot.setRoot(parent);
        } else {
          snapshot.setRoot(databaseManager.reattach(Snapshot.class, parent.getRoot().getId()));
        }
      }
      snapshot = (Snapshot) databaseManager.save(snapshot);
      snapshotByEntity.put(entity, snapshot);
    }
    return snapshot;
  }
}
