/*
 * Sonar, entreprise quality control tool.
 * Copyright (C) 2007-2008 Hortis-GRC SA
 * mailto:be_agile HAT hortis DOT ch
 *
 * Sonar is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * Sonar is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with Sonar; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02
 */
package org.sonar.maven.checkstyle;

import ch.hortis.sonar.model.RuleFailureLevel;
import ch.hortis.sonar.model.RuleFailureParam;
import org.apache.maven.project.MavenProject;
import org.sonar.maven.MeasuresRecorder;
import org.sonar.commons.RulesRepository;
import org.sonar.maven.AbstractCollector;
import org.sonar.maven.XmlParserException;
import org.sonar.maven.XmlReportParser;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import java.io.File;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.List;

public class CheckstyleCollector extends AbstractCollector {
  private XmlReportParser parser;
  private MeasuresRecorder recorder;
  private List<String> sourceDirs;
  private RulesRepository rulesRepository;

  public CheckstyleCollector(MavenProject project, MeasuresRecorder recorder, RulesRepository rulesRepository) {
    File reportFile = findFileFromBuildDirectory(project, "checkstyle-result.xml");
    if (reportFile != null) {
      parser = new XmlReportParser();
      parser.parse(reportFile);
      this.recorder = recorder;
      sourceDirs = project.getCompileSourceRoots();
      this.rulesRepository = rulesRepository;
    }
  }

  public CheckstyleCollector(File xmlFile, String sourceDirectory, MeasuresRecorder recorder, RulesRepository rulesRepository) {
    parser = new XmlReportParser();
    parser.parse(xmlFile);
    this.recorder = recorder;
    this.sourceDirs = new ArrayList<String>();
    this.sourceDirs.add(sourceDirectory);
    this.rulesRepository = rulesRepository;
  }


  public void collect() {
    if (parser == null) {
      return;
    }

    try {
      NodeList files = parser.executeXPathNodeList("/checkstyle/file");
      if (files != null) {
        for (int i = 0; i < files.getLength(); i++) {
          Element element = (Element) files.item(i);
          String checkstyleName = element.getAttribute("name");

          List<Element> errors = parser.getChildElements(element, "error");
          for (Element error : errors) {
            String source = error.getAttribute("source");
            if (rulesRepository.getRuleByPluginKey(source) == null) {
              continue;
            }

            String[] s = getPackageAndClass(checkstyleName, sourceDirs);
            if (s == null) {
              continue;
            }
            String severity = error.getAttribute("severity");
            RuleFailureLevel level = RuleFailureLevel.INFO;
            if ("error".equals(severity)) {
              level = RuleFailureLevel.ERROR;
            } else if ("warning".equals(severity)) {
              level = RuleFailureLevel.WARNING;
            }

            RuleFailureParam lineParam = null, columnParam = null;
            String line = error.getAttribute("line");
            if (line != null && !"".equals(line)) {
              lineParam = new RuleFailureParam("line", parseNumber(line), null);
            }
            String column = error.getAttribute("column");
            if (column != null && !"".equals(column)) {
              columnParam = new RuleFailureParam("column", parseNumber(column), null);
            }

            recorder.createClassRuleFailure(rulesRepository.getRuleByPluginKey(source), s[0], s[1], false, error.getAttribute("message"), level, lineParam, columnParam);
          }
        }
      }
    } catch (ParseException e) {
      throw new XmlParserException(e);
    }
  }
}
